/*{{{  ida.c*/
#include "expr.h"
#include "externs.h"
#include "ida.h"
#include "idafile.h"
#include "lex.h"	/* for 'start_tk', 'stmt_endp []', 'stmt_lineno []' */
#include "output.h"
#include "strmacs.h"
#include "symbol.h"
#include "unit.h"
#include "typedef.h"

Exprlist  *list_stack;		/*  used in 'yyparse()', 'yyerror()',	*/
				/* 'new_exprlist_ref()', 'remove_exprlist_ref()'*/

/********************
-- ALSO TO DO:

-- Doesn't handle actual argument procs yet; e.g. in
	'convert_ida_read_ref_to_actual()', need to consider ND_FUNC & ND_SUBR.

-- ** VERY IMPORTANT **
Check that 'convert_ida_read... ()' and 'install_ida_proc_ref()' only
installs IDA_ACTUAL & IDA_ARG_VAR if approp (i.e. for variables, not params,
fns, etc), and that an IDA_READ record always actually exists when it's changed to the above!

-- MUST HANDLE CHAR*(len) -- O/P THIS AS C[len]{shape}
***********************/

/*{{{  new_exprlist_ref (sp) -- seems ok!*/
new_exprlist_ref (sp)
/*-----------------*/
	Symbol  *sp;		/* may be NULL */
/*----------------------------------------------------------------
	Called by 'yyparse()' upon encountering an ID followed
by a ( [arg-list] ), which may be a function, subroutine, structure
constructor or array reference, or an I/O stmt (or an error!).
	This function adds a record at the head of 'list_stack',
which holds the following information while the argument/subscript
list is processed:
   list_stack->sp:		the proc / array to which the expr list belongs;
   list_stack->lineno:		lineno on which the ref (to 'sp') starts;
   list_stack->nexpr:		# of current expr in the list;
   list_stack->last_expr:	pointer to the current last expr in the list.

The first 2 are required by 'convert_ida_read_ref_to_actual()',
the 3rd is used by 'install_ida_proc_ref()' to obtain the lineno
for the 'IDA_PROC_REF' record, and the last helps 'add_expr ()' build
the arg-list (as it saves traversing the list to find its end every
time a new arg is added).

N.B.  We defer installing an 'IDA_PROC_REF' record for 'sp' until
its arg-list has been parsed and processed (i.e. until the rule
'part_ref' or 'call_stmt' is reduced), since only then can we be
certain whether 'sp' is a non-intrinsic procedure.   E.g. a scalar
CHARACTER object could be either a variable or a function;  which
it is is resolved by its argument list.
----------------------------------------------------------------*/
{
	int  i;
	Exprlist  *list_ref;

	/*------------------------------------------------------*/
	/* add new 'expr list' record to head of 'list_stack'	*/
	/*------------------------------------------------------*/
	list_ref = (Exprlist *) emalloc (sizeof (Exprlist));

	list_ref->next = list_stack;	/* 'list_stack' is a global var! */
	list_stack = list_ref;
	
	list_ref->sp    = sp;
	list_ref->nexpr = 0;
	list_ref->last_expr = (Expr *) NULL;

	/*------------------------------------------------------*/
	/* find lineno of the current position in the stmt	*/
	/* (viz. the lineno on which the 'proc-name' appears).	*/
	/*------------------------------------------------------*/
	for (i = 0;  start_tk > stmt_endp [i];  i++)
		;
	list_ref->lineno = stmt_lineno [i];
}
/*}}}*/
/*{{{  add_expr (ep)*/
add_expr (ep)
/*---------*/
	Expr  *ep;	/* pointer to a 'ND_EXPRLIST' node (a member of an
			   expr-list, e.g an actual arg, subscr expr, etc).
			   May be NULL (for an invalid expr!) */
/*----------------------------------------------------------------
	Adds 'ep' to the *end* of the current expr-list
and updates 'list_stack->last_expr / nexpr' accordingly.

N.B. Since this is *always* invoked when any arg-list or
subscr-list parsed, a record *must* be added to 'list_stack' when
anything with such a list is parsed, even for I/O stmts like READ,
WRITE, OPEN, etc.
----------------------------------------------------------------*/
{
	Expr  *last_expr = list_stack->last_expr;

	if (! ep)
		return;

	if (last_expr)
		attach_expr (last_expr->right, ep);   /* incr 'ep->nlinks' */

	list_stack->nexpr++;
	list_stack->last_expr = ep;
}
/*}}}*/
/*{{{  remove_exprlist_ref ()*/
remove_exprlist_ref ()
/*------------------*/
/*--------------------------------------------------------------*/
/*	Called by 'yyparse()' at the end of an ( [arg-list] ),	*/
/* to remove the 'Exprlist' record at the head of 'list_stack'.	*/
/*--------------------------------------------------------------*/
{
	Exprlist  *listp;

	listp = list_stack;	/* 'list_stack' is a global var! */

	list_stack = listp->next;

	free ((char *) listp);
}
/*}}}*/

#ifdef V_IDA

/*{{{  install_ida_ref (sp, ref) -- seems ok!*/
install_ida_ref (sp, ref)
/*---------------------*/
	Symbol  *sp;
	int  ref;	/* IDA_READ or IDA_WRITE */
/*--------------------------------------------------------------*/ 
/*	Called by 'yyparse()' to install another entry in 'sp's	*/
/* list of references.						*/
/*								*/
/* N.B. We check that 'sp' isn't a named constant, as we don't	*/
/* want to install a ref record for such an id.  If 'sp' occurs	*/
/* in an expression then the caller performs this check (i.e. it*/
/* doesn't call this if 'ep->type' has EX_CONST set), but some	*/
/* erroneous programs (e.g. where a parameter is used as a	*/
/* variable) may slip though this check.			*/
/*--------------------------------------------------------------*/ 
/*{{{  *** THINGS NOT TREATED PROPERLY YET ****/
/*--------------------------------------------------------------
Refs that aren't treated properly yet:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

IDA_WRITE not installed for:
---------------------------
-- internal file in WRITE stmt
-- dummy arg on proc entry (? -- do we want to do this?)
-- IOSTAT=, SIZE= specifiers
-- any args that may be written in INQUIRE stmts.

More generally:
--------------
-- Don't take account of equivalences.
--------------------------------------------------------------*/
/*}}}*/
{
	int  i,  lineno;
	Idref  *refp;

	if (sp->attrs & AT_PARAM) return;
	
	/*------------------------------------------------------*/
	/* find lineno of relevant position in the stmt.	*/
	/*------------------------------------------------------*/
	for (i = 0;  start_tk > stmt_endp [i];  i++)
		;
	lineno = stmt_lineno [i];

	/*------------------------------------------------------*/
	/* create a new 'Idref' record and add it to 'sp's list	*/
	/* of references.					*/
	/*------------------------------------------------------*/
	refp = (Idref *) emalloc (sizeof (Idref));
	refp->lineno = lineno;
	refp->ref    = ref;
	refp->strval = NULL;
	refp->next   = NULL;
	
	if (sp->last_refp)
		sp->last_refp = sp->last_refp->next = refp;
	else
		/*------------------------------*/
		/* the first reference to 'sp'!	*/
		/*------------------------------*/
		sp->first_refp = sp->last_refp = refp;
}
/*}}}*/
/*{{{  convert_ida_read_ref_to_actual (sp) -- seems ok!*/
convert_ida_read_ref_to_actual (sp)
/*--------------------------------*/
	Symbol  *sp;
/*----------------------------------------------------------------
	Called by 'yyparse()' when the rule for 'actual_arg' is
reduced and the following conditions hold:

-- 'actual_arg' is a variable, 'sp' (rather than a constant
	or a read-only expression, for example);  *and*
-- 'actual_arg' is an argument of a non-intrinsic procedure
	(rather than an argument of an intrinsic procedure or
	I/O stmt, etc, or a subscript, or a component of a
	structure constructor).

This function converts 'sp's 'IDA_READ' record into an 'IDA_ACTUAL'
one.
	'sp's 'IDA_READ' record for this reference was installed
when:
	expr :	primary
was reduced, which preceded the reduction that invoked this function, viz:
       (R1211)   actual_arg :	expr
Also, this record is the *last* in 'sp's 'Idref' list.  This
record is kept but converted it to an 'IDA_ACTUAL' record of the form:
	refp->lineno:	lineno       (already given)
	refp->ref:	IDA_ACTUAL | narg
	refp->strval:	"proc-name"
**********************************  N.B.:
-- FOR INTRIN PROCS SHOULD CHECK THE EXPLICIT INTERFACE, AS COULD BE
READ *OR* WRITE REFS!
----------------------------------------------------------------*/
{
	char  *proc_name = list_stack->sp->name;
				/* N.B. 'list_stack' is a global var! */
	Idref  *refp;

	/*------------------------------------------------------*/
	/* Convert 'sp's 'IDA_READ' record -> 'IDA_ACTUAL'	*/
	/*------------------------------------------------------*/
	refp = sp->last_refp;
	refp->ref = (IDA_ACTUAL | (list_stack->nexpr + 1));
				/* +1 because 'add_expr()' (which incre-*/
				/* ments 'narg') hasn't been called yet	*/
	
	refp->strval = (char *) emalloc (strlen (proc_name) + 1);
	strcpy (refp->strval, proc_name);
}
/*}}}*/
/*{{{  convert_intrin_actual_refs_to_read (ep) -- seems ok!*/
convert_intrin_actual_refs_to_read (ep)
/*-----------------------------------*/
	Expr  *ep;	/* a 'ND_FUNC' tree for an intrinsic function ref */
/*----------------------------------------------------------------
	Called by 'yyparse()' when the rule for 'part_ref' is
reduced, and the nature of its 'array_or_func_name' was unknown
previously but is now resolved to be an intrinsic function.
	In this case, the function's arguments have had 'IDA_ACTUAL'
records installed for them (by 'convert_ida_read_ref_to_actual ()'),
but they should actually be 'IDA_READ' refs, since all intrinsic
functions have read-only arguments.  These records are corrected
to 'IDA_READ' ones here.
----------------------------------------------------------------*/
{
	Expr   *arg_ep;
	Idref  *refp;
	Symbol *arg_sp;
	char   *old_strval = ep->val.sp->name;  /* the function name */
	int   narg,  old_ref;

	for (ep = ep->right,  narg = 1;   ep;   ep = ep->right,  narg++) {

		arg_ep = ep->left;	/* since 'ep' is a 'ND_EXPRLIST' node */

		if (! (arg_ep->token == ND_SCALAR ||
		       arg_ep->token == ND_WHOLE_ARRAY ||
		       arg_ep->token == ND_ARRAY_SECT) )
			continue;

		arg_sp = arg_ep->val.sp;	/* the argument variable */

		if (arg_sp->attrs & AT_PARAM)
			return;		/* it isn't a variable argument! */

		/*-----------------------------------------------*/
		/* Search for and change the 'IDA_ACTUAL' record */
		/*-----------------------------------------------*/
		old_ref = (IDA_ACTUAL | narg);
		for (refp = arg_sp->first_refp;  refp;  refp = refp->next) {

			if (refp->ref != old_ref  ||
			    str_mismatch (refp->strval, old_strval))
				continue;

			/*----------------------------*/
			/* Found it -- now change it! */
			/*----------------------------*/
			refp->ref = IDA_READ;
			free (refp->strval);
			refp->strval = (char *) NULL;
			break;		/* exit 'for' loop */
		}
	}
}
/*}}}*/
/*{{{  install_ida_proc_actual_ref (sp) -- seems ok!*/
install_ida_proc_actual_ref (sp)
/*----------------------------*/
	Symbol  *sp;	/* a procedure (external or intrinsic) */
/*----------------------------------------------------------------
	Called by 'yyparse()' when the rule for 'actual_arg' is
reduced and the following conditions hold:

-- 'actual_arg' is a procedure name 'sp' (i.e. this is a procedure
	actual argument);  *and*
-- 'actual_arg' is an argument of a non-intrinsic procedure
	(rather than an argument of an intrinsic procedure or
	I/O stmt, etc, or a subscript, or a component of a
	structure constructor).

This function installs an 'IDA_PROC_ACTUAL' record for 'sp' of the form:
	refp->lineno:	lineno
	refp->ref:	IDA_PROC_ACTUAL | narg
	refp->strval:	"proc-name"
----------------------------------------------------------------*/ 
{
	char  *proc_name = list_stack->sp->name;
				/* N.B. 'list_stack' is a global var! */
	int  i,  lineno;
	Idref  *refp;

	/*------------------------------------------------------*/
	/* find lineno of relevant position in the stmt.	*/
	/*------------------------------------------------------*/
	for (i = 0;  start_tk > stmt_endp [i];  i++)
		;
	lineno = stmt_lineno [i];

	/*------------------------------------------------------*/
	/* create a new 'Idref' record and add it to 'sp's list	*/
	/* of references.					*/
	/*------------------------------------------------------*/
	refp = (Idref *) emalloc (sizeof (Idref));
	refp->lineno = lineno;
	refp->ref = (IDA_PROC_ACTUAL | (list_stack->nexpr + 1));
				/* +1 because 'add_expr()' (which incre-*/
				/* ments 'narg') hasn't been called yet	*/

	refp->strval = (char *) emalloc (strlen (proc_name) + 1);
	strcpy (refp->strval, proc_name);
	refp->next = (Idref *) NULL;
	
	if (sp->last_refp)
		sp->last_refp = sp->last_refp->next = refp;
	else
		/*------------------------------*/
		/* the first reference to 'sp'!	*/
		/*------------------------------*/
		sp->first_refp = sp->last_refp = refp;
}
/*}}}*/
/*{{{  install_ida_proc_ref (ep)  -- seems ok!*/
install_ida_proc_ref (ep)
/*---------------------*/
	Expr  *ep;
/*----------------------------------------------------------------
	Called by 'yyparse()' when the rule for 'part_ref'
(with an arg-list) or 'subr_call' is reduced.  This installs
a list of 'IDA' records for the procedure name (ep->val.sp), viz:
an 'IDA_PROC_REF' for the procedure call, followed by an 'IDA_ARG_xxx'
record for each actual argument (whatever its nature).
	The caller ensures that 'ep' really is a a non-intrinsic
procedure reference, so this needn't be checked.

******** N.B. DOESN'T HANDLE ACTUAL ARG PROCS YET !! **********
----------------------------------------------------------------*/
{
	Expr    *arg_ep;
	Idref   *refp;
	Symbol  *arg_sp,  *proc_sp = ep->val.sp;

	/*{{{  install an 'IDA_PROC_REF' record in 'proc_sp'*/
	/*--------------------------------------------------------------*/
	/* Create an 'IDA_PROC_REF' record and add it to 'proc_sp's	*/
	/* list of references.						*/
	/*								*/
	/* N.B. we do it explicitly rather than by calling		*/
	/* 'install_ida_ref ()' as the latter would install the wrong	*/
	/* 'lineno' (viz. the current one rather than that stored in	*/
	/* 'list_stack').						*/
	/*--------------------------------------------------------------*/
	refp = (Idref *) emalloc (sizeof (Idref));
	
	refp->lineno = list_stack->lineno;	/* 'list_stack' is a global var */
	refp->ref    = IDA_PROC_REF;
	refp->strval = (char *) NULL;
	refp->next   = (Idref *) NULL;
	
	if (proc_sp->last_refp)
		proc_sp->last_refp = proc_sp->last_refp->next = refp;
	else
		/*--------------------------------------*/
		/* the first reference to 'proc_sp'!	*/
		/*--------------------------------------*/
		proc_sp->first_refp = proc_sp->last_refp = refp;
	/*}}}*/

	for (ep = ep->right;  ep;  ep = ep->right) {
		/*---------------------------------------------------------*/
		/* install an 'IDA_ARG_xxx' record for the next actual arg */
		/*---------------------------------------------------------*/
		arg_ep = ep->left;	/* since 'ep' is a 'ND_EXPRLIST' node */
		
		refp = (Idref *) emalloc (sizeof (Idref));
		refp->next = (Idref *) NULL;
		
		switch (arg_ep->token) {
		case  ND_SCALAR :
		case  ND_WHOLE_ARRAY :
		case  ND_ARRAY_SECT :
		case  ND_PROC_ACTUAL_ARG :
			arg_sp = arg_ep->val.sp;
			if (arg_sp->attrs & AT_PARAM)
				goto  expr_arg;
		
			/*{{{  make an 'IDA_ARG_VAR' or 'IDA_ARG_PROC' record*/
			/*--------------------------------------------------------------*/
			/* Make an IDA_ARG_VAR or IDA_ARG_PROC record for 'proc_sp',	*/
			/* of the form:							*/
			/*	refp->lineno:	--   (not needed)			*/
			/*	refp->ref:	IDA_ARG_VAR  or  IDA_ARG_PROC		*/
			/*	refp->strval:	"actual-arg-expr"			*/
			/*								*/
			/* (It's installed after exiting this 'switch').		*/
			/*--------------------------------------------------------------*/
			refp->ref = (arg_ep->token == ND_PROC_ACTUAL_ARG) ? IDA_ARG_PROC : IDA_ARG_VAR;
			
			refp->strval = (char *) emalloc (expr_len (arg_ep, 0L, 0L) + 1);
			write_expr_to_str (refp->strval, arg_ep, 0L, 0L);
								/* copy the arg to a string */
			
			break;
			/*}}}*/
		
		case  ND_ALT_RETURN_LABEL :
			/*{{{  make an 'IDA_ARG_ALT_RET' record*/
			/*--------------------------------------------------------------*/
			/* An alternate return actual argument				*/
			/* -----------------------------------				*/
			/* Make an 'IDA_ARG_ALT_RET' record for 'proc_sp', of the form:	*/
			/*	refp->lineno:	--   (not needed)			*/
			/*	refp->ref:	IDA_ARG_ALT_RETURN			*/
			/*	refp->strval:	NULL					*/
			/*								*/
			/* (It's installed after exiting this 'switch').		*/
			/*--------------------------------------------------------------*/
			refp->ref    = IDA_ARG_ALT_RET;
			refp->strval = (char *) NULL;
				
			break;
			/*}}}*/
		
		default :
		expr_arg :
			/*{{{  make an 'IDA_ARG_type' record*/
			/*--------------------------------------------------------------*/
			/* An expression actual argument				*/
			/* -----------------------------				*/
			/* Make an 'IDA_ARG_type' record for 'proc_sp', of the form:	*/
			/*	refp->lineno:	--   (not needed)			*/
			/*	refp->ref:	IDA_ARG_type				*/
			/*	refp->strval:	NULL					*/
			/*								*/
			/* (It's installed after exiting this 'switch').		*/
			/*--------------------------------------------------------------*/
			switch (arg_ep->type & EX_BASIC_TYPE) {
			case  EX_INTEGER :	refp->ref = IDA_ARG_INTEGER;	break;
			case  EX_REAL :		refp->ref = IDA_ARG_REAL;	break;
			case  EX_DP :		refp->ref = IDA_ARG_DP;		break;
			case  EX_COMPLEX :	refp->ref = IDA_ARG_COMPLEX;	break;
			case  EX_LOGICAL :	refp->ref = IDA_ARG_LOGICAL;	break;
			case  EX_CHARACTER :	refp->ref = IDA_ARG_CHAR;	break;
			case  EX_DERIVED :	refp->ref = IDA_ARG_DERIVED;	break;
			case  EX_NONE :  case  0 :
						refp->ref = IDA_ARG_NONE;	break;
			}
			
			refp->strval = (char *) NULL;
			
			break;
			/*}}}*/
		}
		proc_sp->last_refp = proc_sp->last_refp->next = refp;
	}
}


/*}}}*/

/*{{{  write_ida_info ()*/
write_ida_info ()
/*-------------*/
/*--------------------------------------------------------------*/
/*	Called by 'close_current_unit ()' at the end of an	*/
/* external program unit to write all of its relevant info to 	*/
/* the 'ida' file.						*/
/*								*/
/* N.B. This is called at the end of *every* program unit,	*/
/* regardless of errors.  E.g. if the unit is a procedure, it	*/
/* isn't guaranteed that it's named (although it's erroneous if	*/
/* it isn't).							*/
/*--------------------------------------------------------------*/
{
	Grouplist  *cp;		/* list of common block records */
	Symbol  *init_sp,  *sp;
	Symlist  *idp;		/* list of IDs in each common block */
	
/*** GOT TO DO ERR CHECKING (e.g. disk full!) AS WELL! ****/

/*{{{  code letters for the 'entity kind' field*/
/*-----------------------------------------------------------
entity field (e)
----------------
e = F -- function header  (data type)	--
    S -- subr header			--
    P -- main program header		--

(don't output anything for: block data / module)

    d -- dummy var,	(data type)	-- (used for all bummy args at moment!)
    n -- dummy fn,	(data type)
    o -- dummy subr,
    p -- dummy proc	(unknown proc-type)
    
    c -- common block			--
    g -- global var	(data type)	--
    l -- local var	(data type)

    f -- external fn	(data type)
    s -- external subr
    e -- external proc	(unknown proc-type)
-----------------------------------------------------------*/
/*}}}*/


	/* assume we're at the external-level when this is invoked! */

	/*{{{  output program unit record;  return if block data or module*/
	
	switch (this_unitp->attrs & UNIT_PROC_TYPE) {
	
	case  SUBROUTINE_UNIT :	write_ida_rec ('S', this_unitp->sp);	break;
	
	case  FUNCTION_UNIT :	write_ida_rec ('F', this_unitp->sp);	break;
	
	case  PROGRAM_UNIT :	write_ida_rec ('P', this_unitp->sp);	break;
	
	default :		return;
				/* no output for a block data or module! */
	}
	/*}}}*/

	/*{{{  find 'init_sp' (the next 'sp' after 'prog-unit' name). Return if NULL*/
	/*--------------------------------------------------------------*/
	/* Set 'init_sp' to the ID after the program unit name, or to	*/
	/* the first ID if the program unit is unnamed -- e.g. if it's	*/
	/* an unnamed main program.					*/
	/*								*/
	/* N.B.								*/
	/* (1) The 'prog-unit-name' needn't be the first ID in the	*/
	/* symtab.  E.g., in:						*/
	/*								*/
	/*	TYPE (t) FUNCTION f (x)					*/
	/*								*/
	/* the unit name 'f' is the 2nd ID in the symtab!		*/
	/* (2) The program unit name may, exceptionally, be the *only*	*/
	/* name in the symtab (though the program unit isn't very	*/
	/* useful in that case!), in which case 'init_sp' is NULL. E.g.:*/
	/*								*/
	/*	SUBROUTINE s						*/
	/*	  CALL s						*/
	/*	END							*/
	/*--------------------------------------------------------------*/
	init_sp = this_unitp->first_sp;
	
	if (this_unitp->sp) {
		while (init_sp != this_unitp->sp)
			init_sp = init_sp->nextid;
		init_sp = init_sp->nextid;
	}
	if (! init_sp)
		return;
	/*}}}*/

	/*{{{  output dummy arguments*/
	
	for (sp = init_sp;  sp && sp->attrs & AT_DUMMY;  sp = sp->nextid) {
		if      (sp->attrs & AT_FUNC)		write_ida_rec ('n', sp);
		else if (sp->attrs & AT_SUBR)		write_ida_rec ('o', sp);
		else if (sp->attrs & AT_EXTERNAL)	write_ida_rec ('p', sp);
		else					write_ida_rec ('d', sp);
	}
	/*}}}*/
	/*{{{  output common blocks & their contents*/
	for (cp = common_list;  cp;  cp = cp->next) {
		if ( ! cp->first_id)
			/*----------------------------------------------*/
			/* No contents declared for this common block!	*/
			/* Perhaps it only appeared in a SAVE stmt, or	*/
			/* perhaps it was actually declared empty	*/
			/* (which is an error)!  Either way, don't	*/
			/* output it.					*/
			/*----------------------------------------------*/
			continue;
	
		/*------------------------------------*/
		/* write record for common block name */
		/*------------------------------------*/
		if (cp->sp)
			printf ("c%s \n", cp->sp->name);
		else
			fputs ("c \n", stdout);		/* blank common */
	
		/*------------------------------------------------------*/
		/* write a record for each variable in the common block */
		/*------------------------------------------------------*/
		for (idp = cp->first_id;  idp;  idp = idp->next)
			write_ida_rec ('g', idp->sp);
	}
	/*}}}*/
	/*{{{  output local variables   ******/
	for (sp = init_sp;  sp;  sp = sp->nextid) {
		/*-------------------------------------------*/
		/* skip 'sp' if it isn't a local variable... */
		/*-------------------------------------------*/
		if (sp->attrs & (AT_PARAM | AT_PROCEDURE | AT_NONDATA_ENTITY |
		                 AT_DUMMY | AT_RESULT_VAR | AT_IN_COMMON))
	
			continue;
	
	/***************************************************
	 ORIGINALLY HAD ANOTHER TEST INCLUDED IN THE ABOVE ...
	
		     ||  only_flags_set (sp->attrs, AT_COMMON_NAME) )
	
	BUT THIS EXCLUDED EVERYTHING THAT HAD **NO FLAGS AT ALL** SET -- WHICH
	OCCURS FOR A LOT OF THINGS !
	
	For purpose of this demo -- do a cludge -- if it *is* a common block name,
	and nothing else set -- forget it!!
	***************************************************/
		if ( (sp->attrs & AT_COMMON_NAME)  &&
		      only_flags_set (sp->attrs, AT_COMMON_NAME)  &&
		      ! sp->type  &&  ! sp->arp)
	
			continue;	/* assume it's just a common block name! */
	
		write_ida_rec ('l', sp);
	}
	/*}}}*/
	/*{{{  output procedures*/
	for (sp = init_sp;  sp;  sp = sp->nextid) {
		/*---------------------------------------------------------*/
		/* skip 'sp' if it isn't a non-dummy external procedure... */
		/*---------------------------------------------------------*/
		if ( !(sp->attrs & (AT_FUNC | AT_SUBR | AT_EXTERNAL))  ||
		      (sp->attrs & AT_DUMMY)  ||
		      ((sp->attrs & AT_INFO) && (sp->info.sp->attrs & AT_INTRINSIC)) )
	
			continue;
			
		else if (sp->attrs & AT_FUNC)		write_ida_rec ('f', sp);
		else if (sp->attrs & AT_SUBR)		write_ida_rec ('s', sp);
		else if (sp->attrs & AT_EXTERNAL)	write_ida_rec ('e', sp);
								/* func or subr */
	}
	/*}}}*/
}
/*}}}*/
/*{{{  write_ida_rec (entity, sp)*/
write_ida_rec (entity, sp)
/*----------------------*/
	int  entity;	/* a single character */
	Symbol  *sp;
/*--------------------------------------------------------------*/
/*	Output a complete IDA record for an object 'sp'.	*/
/*--------------------------------------------------------------*/
{
	int     nargs;
	Idref  *refp;
	
	if (!sp) {	/* e.g. for main program, simply write "P \n" */
		printf ("%c%c\n", entity, IDA_FIELD_SEP);
		return;
	}

	/*------------------------------*/
	/* entity 'kind' and name	*/
	/*------------------------------*/
	printf ("%c%s%c", entity, sp->name, IDA_FIELD_SEP);

	/*----------------------*/
	/* data type & shape	*/
	/*----------------------*/
	if (sp->type & TY_TYPE) {
		/*{{{  write data type & shape*/
		/*---------------------------*/
		/* write data type and shape */
		/*---------------------------*/
		switch (sp->type & TY_TYPE) {
		case  TY_INTEGER :	putchar ('i');	break;
		case  TY_REAL :		putchar ('r');	break;
		case  TY_DP :		putchar ('d');	break;
		case  TY_COMPLEX :	putchar ('c');	break;
		case  TY_LOGICAL :	putchar ('l');	break;
		case  TY_CHARACTER :	putchar ('C');	break;	/* forget (* len) for now! */
		case  TY_DERIVED :	putchar ('t');	break;	/* shouldn't happen for F77! */
		}
		
		if (sp->arp) {
			/*------------------------------*/
			/* Write shape  (*not* lb:ub)	*/
			/*------------------------------*/
			Expr  **npp;
			int  dim,  rank = (sp->arp->attrs & RANK_MASK);
		
			for (dim = 0, npp = sp->arp->n;   dim < rank;   dim++, npp++) {
				dim ? putchar (',') : putchar ('(');
				write_expr (stdout, *npp, 0L, 0L);
			}
			putchar (')');
		}
		putchar (IDA_FIELD_SEP);
		
		/*}}}*/
	}

	/*----------------------*/
	/* usage records	*/
	/*----------------------*/
	for (refp = sp->first_refp;  refp;   ) {
		/*{{{  write next usage record*/
		/*--------------------------------------------------------------*/
		/* write a usage record of the form:				*/
		/*								*/
		/*	lineno  usage+  ' '					*/
		/*								*/
		/* (without any space between the grammar symbols!)		*/
		/*	If an object appears several times on the same line,	*/
		/* there is a separate 'usage field' for each one.  E.g.:	*/
		/*								*/
		/* 22:		X = X + FN (X)					*/
		/*								*/
		/* results in usage fields "22w 22r 22a1FN" for 'X'.		*/
		/*--------------------------------------------------------------*/
		printf ("%d", refp->lineno);
		
		switch (refp->ref & IDA_REF_MASK) {
		case  IDA_READ :	/* a READ ref */
			putchar ('r');
			refp = refp->next;
			break;
		
		case  IDA_WRITE :	/* a WRITE ref	*/
			putchar ('w');
			refp = refp->next;
			break;
		
		case  IDA_ACTUAL :	/* a variable passed as an actual argument */
			printf  ("a%d%s",
			         (refp->ref & IDA_NARG_MASK),	/* # (arg) */
			          refp->strval);		/* proc name */
			refp = refp->next;
			break;
		
		case  IDA_PROC_ACTUAL :	/* a procedure passed as an actual argument */
			printf  ("b%d%s",
			         (refp->ref & IDA_NARG_MASK),	/* # (arg) */
			          refp->strval);		/* proc name */
			refp = refp->next;
			break;
		
		case  IDA_PROC_REF :	/* a proc reference with its argument list */
			putchar ('p');
			/*{{{  output the argument list.  (This advances 'refp')*/
			/*------------------------------------------------------*/
			/* Output the argument list in the form:  (arg,arg,...)	*/
			/*------------------------------------------------------*/
			putchar ('(');
			for (refp = refp->next,  nargs = 0;
			     refp  &&  (refp->ref & IDA_ARG_MASK);
			     refp = refp->next,  nargs++) {
			
				if (nargs)
					putchar (',');
			
				switch (refp->ref & IDA_ARG_MASK) {
				case  IDA_ARG_VAR :				/* arg is a variable */
					printf ("V%c%s%c",
					         IDA_OPEN_PAREN, refp->strval, IDA_CLOSE_PAREN);
					break;
			
				case  IDA_ARG_PROC :				/* arg is a proc name */
					printf ("P%c%s%c",
					         IDA_OPEN_PAREN, refp->strval, IDA_CLOSE_PAREN);
					break;
			
				case  IDA_ARG_ALT_RET :	putchar ('A');	break;
				case  IDA_ARG_INTEGER :	putchar ('I');	break;	/* arg is an int expr */
				case  IDA_ARG_REAL :	putchar ('R');	break;	/* etc...   */
				case  IDA_ARG_DP :	putchar ('D');	break;
				case  IDA_ARG_COMPLEX :	putchar ('C');	break;
				case  IDA_ARG_LOGICAL :	putchar ('L');	break;
				case  IDA_ARG_CHAR :	putchar ('X');	break;
				case  IDA_ARG_DERIVED :	putchar ('Y');	break;
				case  IDA_ARG_NONE :	putchar ('N');	break;
				}
			}
			putchar (')');
			/*}}}*/
			break;
		}
		
		putchar (IDA_FIELD_SEP);
		/*}}}*/
	}

	putchar ('\n');
}
/*}}}*/

/*{{{  delete_ida_recs (refp)*/
delete_ida_recs (refp)
/*------------------*/
	Idref  *refp;	/* initialised to 'sp->first_refp', for some 'sp' */
{
	Idref  *next_refp;

	for ( ;  refp;  refp = next_refp) {
		next_refp = refp->next;

		if (refp->strval)
			free (refp->strval);
		free ((char *) refp);
	}
}
/*}}}*/
#endif
/*}}}*/
